DATASET_FIDDLER_ID = '__fiddler_id'
ONE_LINE_PRINT = '¡í€'
TIMESTAMP_FORMAT = '%Y-%m-%d %H:%M:%S.%f'
MAX_ID_LEN = 30


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDCOLOR = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
